-- Deleted contexts data
CREATE TABLE context_deleted ( id TEXT PRIMARY KEY,
    predefined_id INTEGER DEFAULT 0,
    name TEXT COLLATE NOCASE NOT NULL,
    memo TEXT COLLATE NOCASE DEFAULT NULL,
    longitude REAL DEFAULT 0,
    latitude REAL DEFAULT 0,
    altitude REAL DEFAULT 0,
    timestamp INTEGER DEFAULT NULL,
    on_hold INTEGER DEFAULT 0,
    locked INTEGER DEFAULT 0
);

-- Deleted projects data
CREATE TABLE project_deleted ( id TEXT PRIMARY KEY,
    predefined_id INTEGER DEFAULT 0,
    name TEXT COLLATE NOCASE NOT NULL,
    memo TEXT COLLATE NOCASE DEFAULT NULL,
    -- Foreign key
    parent TEXT DEFAULT NULL,
    start INTEGER DEFAULT NULL,
    due INTEGER DEFAULT NULL,
    -- Foreign key
    context TEXT DEFAULT NULL,
    folder INTEGER DEFAULT 0,
    type INTEGER DEFAULT 0,
    timestamp INTEGER DEFAULT NULL,
    completed INTEGER DEFAULT 0,
    on_hold INTEGER DEFAULT 0,
    locked INTEGER DEFAULT 0
);

-- Deleted tasks data
CREATE TABLE task_deleted ( id TEXT PRIMARY KEY,
    predefined_id INTEGER DEFAULT 0,
    name TEXT COLLATE NOCASE NOT NULL,
    memo TEXT COLLATE NOCASE DEFAULT NULL,
    -- Foreign key
    parent TEXT NOT NULL,
    -- Foreign key
    context TEXT DEFAULT NULL,
    start INTEGER DEFAULT NULL,
    due INTEGER DEFAULT NULL,
    timestamp INTEGER DEFAULT NULL,
    completed INTEGER DEFAULT 0,
    on_hold INTEGER DEFAULT 0,
    locked INTEGER DEFAULT 0
);

-- Update triggers
-- Task table foreign keys insert constraint
DROP TRIGGER task_insert;

CREATE TRIGGER task_insert
BEFORE INSERT ON task
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL AND (SELECT id FROM project_deleted WHERE id = NEW.parent)
            THEN RAISE (ABORT, 'TASK_PARENT_FK_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT id FROM context WHERE id = NEW.context) IS NULL AND (SELECT id FROM context_deleted WHERE id = NEW.context) IS NULL AND NEW.context IS NOT NULL
            THEN RAISE (ABORT, 'TASK_CONTEXT_FK_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN NEW.predefined_id <> 0 AND (SELECT COUNT(*) FROM task WHERE predefined_id = NEW.predefined_id) > 0
            THEN RAISE (IGNORE)
    END;
END;

CREATE TRIGGER task_deleted_insert
BEFORE INSERT ON task_deleted
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL AND (SELECT id FROM project_deleted WHERE id = NEW.parent)
            THEN RAISE (ABORT, 'TASK_PARENT_FK_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT id FROM context WHERE id = NEW.context) IS NULL AND (SELECT id FROM context_deleted WHERE id = NEW.context) IS NULL AND NEW.context IS NOT NULL
            THEN RAISE (ABORT, 'TASK_CONTEXT_FK_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN NEW.predefined_id <> 0 AND (SELECT COUNT(*) FROM task WHERE predefined_id = NEW.predefined_id) > 0
            THEN RAISE (IGNORE)
    END;
END;

-- Task table parent FK update constraint
DROP TRIGGER task_parent_update;

CREATE TRIGGER task_parent_update
BEFORE UPDATE OF parent ON task
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL AND (SELECT id FROM project_deleted WHERE id = NEW.parent) IS NULL
            THEN RAISE (ABORT, 'TASK_PARENT_FK_UPDATE_VIOLATION')
    END;
END;

CREATE TRIGGER task_deleted_parent_update
BEFORE UPDATE OF parent ON task_deleted
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL AND (SELECT id FROM project_deleted WHERE id = NEW.parent) IS NULL
            THEN RAISE (ABORT, 'TASK_PARENT_FK_UPDATE_VIOLATION')
    END;
END;

-- Project table constraints
DROP TRIGGER project_insert;

CREATE TRIGGER project_insert
BEFORE INSERT ON project
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL AND (SELECT id FROM project_deleted WHERE id = NEW.parent) IS NULL AND NEW.parent IS NOT NULL
            THEN RAISE (ABORT, 'PROJECT_PARENT_INSERT_FIRST_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT id FROM context WHERE id = NEW.context) IS NULL AND (SELECT id FROM context_deleted WHERE id = NEW.context) IS NULL AND NEW.context IS NOT NULL
            THEN RAISE (ABORT, 'PROJECT_CONTEXT_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN NEW.id = NEW.parent
            THEN RAISE (ABORT, 'PROJECT_PARENT_INSERT_SECOND_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT COUNT(*) FROM project WHERE type = 1) >= 1 AND NEW.type = 1
            THEN RAISE (ABORT, 'INBOX_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT COUNT(*) FROM project WHERE type = 2) >= 1 AND NEW.type = 2
            THEN RAISE (ABORT, 'SINGLE_ACTIONS_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT type FROM project WHERE id = NEW.parent) = 1
            THEN RAISE (ABORT, 'PROJECT_PARENT_IS_OF_INVALID_TYPE_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT type FROM project WHERE id = NEW.parent) = 2
            THEN RAISE (ABORT, 'PROJECT_PARENT_IS_OF_INVALID_TYPE_VIOLATION')
    END;

    SELECT CASE
        WHEN NEW.predefined_id <> 0 AND (SELECT COUNT(*) FROM project WHERE predefined_id = NEW.predefined_id) > 0
            THEN RAISE (IGNORE)
    END;
END;

CREATE TRIGGER project_deleted_insert
BEFORE INSERT ON project_deleted
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL AND (SELECT id FROM project_deleted WHERE id = NEW.parent) IS NULL AND NEW.parent IS NOT NULL
            THEN RAISE (ABORT, 'PROJECT_PARENT_INSERT_FIRST_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT id FROM context WHERE id = NEW.context) IS NULL AND (SELECT id FROM context_deleted WHERE id = NEW.context) IS NULL AND NEW.context IS NOT NULL
            THEN RAISE (ABORT, 'PROJECT_CONTEXT_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN NEW.id = NEW.parent
            THEN RAISE (ABORT, 'PROJECT_PARENT_INSERT_SECOND_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT COUNT(*) FROM project WHERE type = 1) >= 1 AND NEW.type = 1
            THEN RAISE (ABORT, 'INBOX_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT COUNT(*) FROM project WHERE type = 2) >= 1 AND NEW.type = 2
            THEN RAISE (ABORT, 'SINGLE_ACTIONS_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT type FROM project WHERE id = NEW.parent) = 1
            THEN RAISE (ABORT, 'PROJECT_PARENT_IS_OF_INVALID_TYPE_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT type FROM project WHERE id = NEW.parent) = 2
            THEN RAISE (ABORT, 'PROJECT_PARENT_IS_OF_INVALID_TYPE_VIOLATION')
    END;

    SELECT CASE
        WHEN NEW.predefined_id <> 0 AND (SELECT COUNT(*) FROM project WHERE predefined_id = NEW.predefined_id) > 0
            THEN RAISE (IGNORE)
    END;
END;

-- Project table parent field constraint
DROP TRIGGER project_parent_update;

CREATE TRIGGER project_parent_update
BEFORE UPDATE OF parent ON project
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL AND (SELECT id FROM project_deleted WHERE id = NEW.parent) IS NULL AND NEW.parent IS NOT NULL OR NEW.id = NEW.parent
            THEN RAISE (ABORT, 'PROJECT_PARENT_UPDATE_VIOLATION')
    END;
END;

CREATE TRIGGER project_deleted_parent_update
BEFORE UPDATE OF parent ON project_deleted
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL AND (SELECT id FROM project_deleted WHERE id = NEW.parent) IS NULL AND NEW.parent IS NOT NULL OR NEW.id = NEW.parent
            THEN RAISE (ABORT, 'PROJECT_PARENT_UPDATE_VIOLATION')
    END;
END;