/* Tables */
/* By default foreign keys support is not enabled on Symbian, so we implement constraints manually */

-- Database schema migration history
CREATE TABLE migration ( id INTEGER PRIMARY KEY,
    major_version INTEGER DEFAULT 0,
    minor_version INTEGER DEFAULT 0,
    file_number INTEGER DEFAULT 0,
    comment TEXT DEFAULT NULL,
    date_applied INTEGER NOT NULL,
    UNIQUE(file_number)
);

-- Server synchronization related data
CREATE TABLE server ( id INTEGER PRIMARY KEY,
    major_version INTEGER DEFAULT 0,
    minor_version INTEGER DEFAULT 0,
    login TEXT DEFAULT NULL,
    prev_login TEXT DEFAULT NULL,
    password TEXT DEFAULT NULL,
    revision INTEGER DEFAULT 0,
    authorized INTEGER DEFAULT 0,
    time_offset INTEGER DEFAULT 0
);

-- Context data
CREATE TABLE context ( id TEXT PRIMARY KEY,
    predefined_id INTEGER DEFAULT 0,
    name TEXT COLLATE NOCASE NOT NULL,
    memo TEXT COLLATE NOCASE DEFAULT NULL,
    longitude REAL DEFAULT 0,
    latitude REAL DEFAULT 0,
    altitude REAL DEFAULT 0,
    timestamp INTEGER DEFAULT NULL,
    on_hold INTEGER DEFAULT 0,
    locked INTEGER DEFAULT 0
);

-- Project data
CREATE TABLE project ( id TEXT PRIMARY KEY,
    predefined_id INTEGER DEFAULT 0,
    name TEXT COLLATE NOCASE NOT NULL,
    memo TEXT COLLATE NOCASE DEFAULT NULL,
    -- Foreign key
    parent TEXT DEFAULT NULL,
    start INTEGER DEFAULT NULL,
    due INTEGER DEFAULT NULL,
    -- Foreign key
    context TEXT DEFAULT NULL,
    folder INTEGER DEFAULT 0,
    type INTEGER DEFAULT 0,
    timestamp INTEGER DEFAULT NULL,
    completed INTEGER DEFAULT 0,
    on_hold INTEGER DEFAULT 0,
    locked INTEGER DEFAULT 0
);

-- Task data
CREATE TABLE task ( id TEXT PRIMARY KEY,
    predefined_id INTEGER DEFAULT 0,
    name TEXT COLLATE NOCASE NOT NULL,
    memo TEXT COLLATE NOCASE DEFAULT NULL,
    -- Foreign key
    parent TEXT NOT NULL,
    -- Foreign key
    context TEXT DEFAULT NULL,
    start INTEGER DEFAULT NULL,
    due INTEGER DEFAULT NULL,
    timestamp INTEGER DEFAULT NULL,
    completed INTEGER DEFAULT 0,
    on_hold INTEGER DEFAULT 0,
    locked INTEGER DEFAULT 0
);

-- List of changes to be uploaded to server
CREATE TABLE sync_queue ( id INTEGER PRIMARY KEY,
    uuid TEXT NOT NULL,
    predefined_id INTEGER DEFAULT 0,
    parent TEXT DEFAULT NULL,
    action TEXT NOT NULL,
    timestamp INTEGER DEFAULT NULL,
    UNIQUE(uuid),
    UNIQUE(uuid, action)
);


/* Triggers */

------------------------------------------------------------------------------------
-- Insertion triggers
------------------------------------------------------------------------------------

-- Sync list table timestamp trigger (make sure that timestamp is valid)
CREATE TRIGGER sync_queue_insert_valid_timestamp
AFTER INSERT ON sync_queue
WHEN NEW.timestamp IS NULL OR NEW.timestamp <= 0
BEGIN
    UPDATE sync_queue SET timestamp = strftime('%s','now') WHERE id = NEW.id;
END;


-- Context table timestamp trigger (make sure that timestamp is valid)
CREATE TRIGGER context_insert_valid_timestamp
AFTER INSERT ON context
WHEN NEW.timestamp IS NULL OR NEW.timestamp <= 0
BEGIN
    UPDATE context SET timestamp = strftime('%s','now') WHERE id = NEW.id;
END;

-- Context table constraints
CREATE TRIGGER context_insert
BEFORE INSERT ON context
BEGIN
    SELECT CASE
        WHEN NEW.predefined_id <> 0 AND (SELECT COUNT(*) FROM project WHERE predefined_id = NEW.predefined_id) > 0
            THEN RAISE (IGNORE)
    END;
END;

-- Project table constraints
CREATE TRIGGER project_insert
BEFORE INSERT ON project
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL AND NEW.parent IS NOT NULL
            THEN RAISE (ABORT, 'PROJECT_PARENT_INSERT_FIRST_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT id FROM context WHERE id = NEW.context) IS NULL AND NEW.context IS NOT NULL
            THEN RAISE (ABORT, 'PROJECT_CONTEXT_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN NEW.id = NEW.parent
            THEN RAISE (ABORT, 'PROJECT_PARENT_INSERT_SECOND_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT COUNT(*) FROM project WHERE type = 1) >= 1 AND NEW.type = 1
            THEN RAISE (ABORT, 'INBOX_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT COUNT(*) FROM project WHERE type = 2) >= 1 AND NEW.type = 2
            THEN RAISE (ABORT, 'SINGLE_ACTIONS_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT type FROM project WHERE id = NEW.parent) = 1
            THEN RAISE (ABORT, 'PROJECT_PARENT_IS_OF_INVALID_TYPE_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT type FROM project WHERE id = NEW.parent) = 2
            THEN RAISE (ABORT, 'PROJECT_PARENT_IS_OF_INVALID_TYPE_VIOLATION')
    END;

    SELECT CASE
        WHEN NEW.predefined_id <> 0 AND (SELECT COUNT(*) FROM project WHERE predefined_id = NEW.predefined_id) > 0
            THEN RAISE (IGNORE)
    END;
END;

-- Project time trigger
CREATE TRIGGER project_insert_time
AFTER INSERT ON project
WHEN NEW.start > NEW.due AND NEW.due IS NOT NULL AND NEW.due > 0
BEGIN
    UPDATE project SET start = NULL WHERE id = NEW.id;
END;

-- Project table timestamp trigger (make sure that timestamp is valid)
CREATE TRIGGER project_insert_valid_timestamp
AFTER INSERT ON project
WHEN NEW.timestamp IS NULL OR NEW.timestamp <= 0
BEGIN
    UPDATE project SET timestamp = strftime('%s','now') WHERE id = NEW.id;
END;

-- Task table foreign keys insert constraint
CREATE TRIGGER task_insert
BEFORE INSERT ON task
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL
            THEN RAISE (ABORT, 'TASK_PARENT_FK_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT id FROM context WHERE id = NEW.context) IS NULL AND NEW.context IS NOT NULL
            THEN RAISE (ABORT, 'TASK_CONTEXT_FK_INSERT_VIOLATION')
    END;

    SELECT CASE
        WHEN NEW.predefined_id <> 0 AND (SELECT COUNT(*) FROM task WHERE predefined_id = NEW.predefined_id) > 0
            THEN RAISE (IGNORE)
    END;
END;

-- Task table timestamp trigger (make sure that timestamp is valid)
CREATE TRIGGER task_insert_valid_timestamp
AFTER INSERT ON task
WHEN NEW.timestamp IS NULL OR NEW.timestamp <= 0
BEGIN
    UPDATE task SET timestamp = strftime('%s','now') WHERE id = NEW.id;
END;

-- Task time trigger
CREATE TRIGGER task_insert_time
AFTER INSERT ON task
WHEN NEW.start > NEW.due AND NEW.due IS NOT NULL AND NEW.due > 0
BEGIN
    UPDATE task SET start = NULL WHERE id = NEW.id;
END;

-- Creation sync queue constraint
CREATE TRIGGER sync_queue_insert
BEFORE INSERT ON sync_queue
BEGIN
    -- Ignore silently any data duplication
    SELECT CASE
        WHEN (SELECT COUNT(*) FROM sync_queue WHERE uuid = NEW.uuid AND action = NEW.action > 0)
            THEN RAISE (IGNORE)
    END;
END;

-- Server table records count constraint
CREATE TRIGGER server_insert
BEFORE INSERT ON server
BEGIN
    SELECT CASE
        WHEN (SELECT COUNT(id) FROM server) > 0
            THEN RAISE (ABORT, 'SERVER_RECORDS_COUNT_VIOLATION')
    END;
END;

------------------------------------------------------------------------------------
-- Update triggers
------------------------------------------------------------------------------------

-- Prevent updating of the locked context records
CREATE TRIGGER context_lock
BEFORE UPDATE OF id, predefined_id, name, memo, longitude, latitude, altitude ON context
WHEN OLD.locked <> 0
BEGIN
    SELECT RAISE (ABORT, 'CONTEXT_LOCK_VIOLATION');
END;

-- Context table id field constraint
CREATE TRIGGER context_uuid_update
AFTER UPDATE OF id ON context
BEGIN
    UPDATE task SET context = NEW.id WHERE context = OLD.id;
    UPDATE project SET context = NEW.id WHERE context = OLD.id;
END;

-- Project table parent field constraint
CREATE TRIGGER project_parent_update
BEFORE UPDATE OF parent ON project
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL AND NEW.parent IS NOT NULL OR NEW.id = NEW.parent
            THEN RAISE (ABORT, 'PROJECT_PARENT_UPDATE_VIOLATION')
    END;
END;

-- Project table time field constraint
CREATE TRIGGER project_time_update
AFTER UPDATE OF start, due ON project
WHEN NEW.due < NEW.start AND NEW.due IS NOT NULL
BEGIN
    UPDATE project SET start = NULL WHERE id = NEW.id;
END;

-- Project table default context field constraint
CREATE TRIGGER project_context_update
BEFORE UPDATE OF context ON project
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM context WHERE id = NEW.context) IS NULL AND NEW.context IS NOT NULL
            THEN RAISE (ABORT, 'PROJECT_CONTEXT_UPDATE_VIOLATION')
    END;
END;

-- Project table type constraint
CREATE TRIGGER project_type_update
BEFORE UPDATE OF type ON project
BEGIN
    SELECT CASE
        WHEN (SELECT COUNT(*) FROM project WHERE type = 1) >= 1 AND NEW.type = 1
            THEN RAISE (ABORT, 'TYPE_UPDATE_VIOLATION')
    END;

    SELECT CASE
        WHEN (SELECT COUNT(*) FROM project WHERE type = 2) >= 1 AND NEW.type = 2
            THEN RAISE (ABORT, 'TYPE_UPDATE_VIOLATION')
    END;
END;

-- Project table type update trigger
CREATE TRIGGER project_uuid_update
AFTER UPDATE OF id ON project
BEGIN
    UPDATE task SET parent = NEW.id WHERE parent = OLD.id;
    UPDATE project SET parent = NEW.id WHERE parent = OLD.id;
END;

-- Prevent updating of the locked project records
CREATE TRIGGER project_lock
BEFORE UPDATE OF id, predefined_id, name, memo, parent, type, folder ON project
WHEN OLD.locked <> 0
BEGIN
    SELECT RAISE (ABORT, 'PROJECT_LOCK_VIOLATION');
END;

-- Task table parent FK update constraint
CREATE TRIGGER task_parent_update
BEFORE UPDATE OF parent ON task
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM project WHERE id = NEW.parent) IS NULL
            THEN RAISE (ABORT, 'TASK_PARENT_FK_UPDATE_VIOLATION')
    END;
END;

-- Task table time field constraint
CREATE TRIGGER task_time_update
AFTER UPDATE OF start, due ON task
WHEN NEW.due < NEW.start AND NEW.due IS NOT NULL
BEGIN
    UPDATE task SET start = NULL WHERE id = NEW.id;
END;

-- Task table completion state constraint
CREATE TRIGGER task_completion_update
BEFORE UPDATE OF completed ON task
BEGIN
    SELECT CASE
        WHEN (SELECT completed FROM project WHERE id = NEW.parent) <> 0 AND NEW.completed = 0
            THEN RAISE (ABORT, 'TASK_COMPLETION_UPDATE_VIOLATION')
    END;
END;

-- Task table context FK update constraint
CREATE TRIGGER task_context_update
BEFORE UPDATE OF context ON task
BEGIN
    SELECT CASE
        WHEN (SELECT id FROM context WHERE id = NEW.context) IS NULL AND NEW.context IS NOT NULL
            THEN RAISE (ABORT, 'TASK_PROJECT_FK_UPDATE_VIOLATION')
    END;
END;

-- Prevent updating of the locked task records
CREATE TRIGGER task_lock
BEFORE UPDATE OF id, predefined_id, name, memo, parent, context, start, due ON task
WHEN OLD.locked <> 0
BEGIN
    SELECT RAISE (ABORT, 'TASK_LOCK_VIOLATION');
END;

-- Server table revision value constraint violation (new revision couldn't be less than the previous one)
CREATE TRIGGER server_revision_constraint
BEFORE UPDATE OF revision ON server
WHEN NEW.revision < OLD.revision AND NEW.revision > 0
BEGIN
    SELECT RAISE (ABORT, 'SERVER_REVISION_CONSTRAINT_VIOLATION');
END;

------------------------------------------------------------------------------------
-- Deletion triggers
------------------------------------------------------------------------------------

-- Prevent deletion of the locked context records
CREATE TRIGGER context_delete_lock
BEFORE DELETE ON context
WHEN OLD.locked <> 0
BEGIN
    SELECT RAISE (ABORT, 'CONTEXT_LOCK_VIOLATION');
END;

-- Prevent deletion of the locked project records
CREATE TRIGGER project_delete_lock
BEFORE DELETE ON project
WHEN OLD.locked <> 0
BEGIN
    SELECT RAISE (ABORT, 'PROJECT_LOCK_VIOLATION');
END;

-- Prevent deletion of the locked task records
CREATE TRIGGER task_delete_lock
BEFORE DELETE ON task
WHEN OLD.locked <> 0
BEGIN
    SELECT RAISE (ABORT, 'TASK_LOCK_VIOLATION');
END;


/* Insert required data */
INSERT INTO migration (major_version, minor_version, file_number, comment, date_applied) VALUES(0, 1, 0, 'Baseline', strftime('%s','now'));
INSERT INTO server (major_version, minor_version, revision) VALUES(0, 3, 0);
